import decimal
import fractions
from typing import Union, Tuple, Optional

from ._framerate import Framerate, FramerateSource
from ._premiere_ticks import PremiereTicks
from ._consts import (
    _PPRO_TICKS_PER_SECOND,
    _FRAMES_PER_FOOT,
    _SECONDS_PER_MINUTE,
    _SECONDS_PER_HOUR,
)
from ._timecode_sections import TimecodeSections
from ._timecode_parsers import _TimecodeParseSource, _parse, _rational_to_frames
from ._timecode_dropframe import _frame_num_to_drop_frame_num


TimecodeSource = Union[_TimecodeParseSource, "Timecode"]
"""
TimecodeSource is the types of source values a timecode can be created from. See
documentation for details on how different value types as converted.
"""


class Timecode:
    def __init__(
        self,
        src: TimecodeSource,
        *,
        rate: Optional[FramerateSource],
    ) -> None:
        """
        Timecode represents the frame at a particular time in a video.

        :param src: The source value to parse. The values type determines how it is
            interpreted:

            - ``vtc.Timecode``: used the ``vtc.Timecode.rational`` seconds value to
              construct a new timecode.

            - ``string``: etiher a timecode, runtime, or feet+frames string. Runtime
              strings must have a fractal seconds place, i.e. '01:00:00.0'.

            - ``int``: frame count.

            - ``decimal.Decimal``: seconds.

            - ``float``: seconds.

            - ``fractions.Fraction``: seconds.

            - ``vtc.PremiereTicks``: Adobe Premiere Pro ticks value.

        :param rate: The framerate to use for this timecode. May be any value which
            can be passed to the constructor of :class:`Framerate` or a
            :class:`Framerate`. May only be ``None`` if ``src`` is  a :class:`Framerate`
            instance.

        :returns: A newly constructed :class:`Timecode` value.

        :raises ValueError: If a string value passed to the constructor is not a valid
            timecode, runtime, or feet+frames, or no rate value can be found.

        :raises TypeError: If passed an unsupported type.
        """
        if isinstance(src, Timecode):
            if rate is not None:
                raise ValueError(
                    "rate must be None if src is vtc.Timecode. To rebase Timecode, use"
                    "vtc.Timecode.rebase",
                )
            rate = src.rate
            src = src.rational

        if rate is None:
            raise ValueError(
                "rate must be set for all Timecode src types except vtc.Timecode",
            )

        self._rate: Framerate = Framerate(rate)
        self._value: fractions.Fraction = _parse(src, self._rate)

    def __repr__(self) -> str:
        """__repr__ prints a timecode as [01:00:00:00 @ [23.98 NTSC]]"""
        return f"[{self.timecode} @ {repr(self._rate)}]"

    def __str__(self) -> str:
        """stringing a Timecode returns the timecode value (ex: 01:00:00:00)"""
        return self.timecode

    def __eq__(self, other: object) -> bool:
        """
        Two timecodes are equal if their rational times are equal. This means that
        even if their framerates are different, two timecodes that represent
        1 hour of time exactly will be equal.

        Note that this is NOT related to actual timecode presentation. An 23.98 NTSC
        timecode of 01:00:00 will NOT be equal to a non-NTSC 23.98 timecode of
        01:00:00 because the NTSC timecode represents a rational time of slightly
        less than 1 hour.
        """
        if not isinstance(
            other, (Timecode, str, int, float, fractions.Fraction, decimal.Decimal)
        ):
            return NotImplemented

        other_tc = _coerce_other(other, self._rate)
        return self._value == other_tc._value

    def __lt__(self, other: TimecodeSource) -> bool:
        other_tc = _coerce_other(other, self._rate)
        return self._value < other_tc._value

    def __le__(self, other: TimecodeSource) -> bool:
        other_tc = _coerce_other(other, self._rate)
        return self._value <= other_tc._value

    def __gt__(self, other: TimecodeSource) -> bool:
        other_tc = _coerce_other(other, self._rate)
        return self._value > other_tc._value

    def __ge__(self, other: TimecodeSource) -> bool:
        other_tc = _coerce_other(other, self._rate)
        return self._value >= other_tc._value

    def __add__(self, other: TimecodeSource) -> "Timecode":
        other_tc = _coerce_other(other, self._rate)
        frac_value = self._value + other_tc._value
        return Timecode(frac_value, rate=self._rate)

    def __sub__(self, other: TimecodeSource) -> "Timecode":
        other_tc = _coerce_other(other, self._rate)
        frac_value = self._value - other_tc._value
        return Timecode(frac_value, rate=self._rate)

    def __mul__(
        self,
        other: Union[int, float, fractions.Fraction, decimal.Decimal],
    ) -> "Timecode":
        # Return a new timecode with the multiplication applied.
        return Timecode(round(self.frames * other), rate=self._rate)

    def __truediv__(
        self,
        other: Union[int, float, fractions.Fraction, decimal.Decimal],
    ) -> "Timecode":
        # Return a new timecode with the multiplication applied.
        return Timecode(round(self.frames / other), rate=self._rate)

    def __floordiv__(
        self,
        other: Union[int, float, fractions.Fraction, decimal.Decimal],
    ) -> "Timecode":
        # Return a new timecode with the multiplication applied.
        return Timecode(int(self.frames // other), rate=self._rate)

    def __mod__(
        self,
        other: Union[int, float, fractions.Fraction, decimal.Decimal],
    ) -> "Timecode":
        return Timecode(int(self.frames % other), rate=self._rate)

    def __divmod__(
        self,
        other: Union[int, float, fractions.Fraction, decimal.Decimal],
    ) -> Tuple["Timecode", "Timecode"]:
        dividend, modulo = divmod(self.frames, other)
        return (
            Timecode(int(dividend), rate=self._rate),
            Timecode(int(modulo), rate=self._rate),
        )

    def __neg__(self) -> "Timecode":
        return Timecode(-self._value, rate=self._rate)

    def __abs__(self) -> "Timecode":
        return Timecode(abs(self._value), rate=self._rate)

    @property
    def rate(self) -> Framerate:
        """rate is the framerate at which this timecode is being interpreted"""
        return self._rate

    @property
    def rational(self) -> fractions.Fraction:
        """
        frac is a rational (fraction) representation of number of seconds this timecode
        represents.
        """
        return self._value

    @property
    def sections(self) -> TimecodeSections:
        """
        sections returns the sections of a timecode as ints for callers to
        format/work on as desired.
        """
        rate = self._rate
        frames_number = abs(self.frames)

        if self._rate.dropframe:
            # We need to do an adjustment for drop-frame timecode
            frames_number = _frame_num_to_drop_frame_num(
                frames_number,
                self.rate.timebase,
            )

        timebase = rate.timebase

        hours, frames = divmod(frames_number, timebase * _SECONDS_PER_HOUR)
        minutes, frames = divmod(frames, timebase * _SECONDS_PER_MINUTE)
        seconds, frames = divmod(frames, timebase)

        return TimecodeSections(
            # If our value is less than 0, this is a negative value.
            negative=self._value < 0,
            hours=hours,
            minutes=minutes,
            seconds=seconds,
            frames=round(frames),
        )

    @property
    def timecode(self) -> str:
        """
        timecode returns the formatted SMPTE timecode: (ex: 01:00:00:00).
        """
        sections = self.sections

        frames_sep = ":"
        if self._rate.dropframe:
            frames_sep = ";"

        timecode = (
            f"{str(sections.hours).zfill(2)}"
            f":{str(sections.minutes).zfill(2)}"
            f":{str(sections.seconds).zfill(2)}"
            f"{frames_sep}{str(sections.frames).zfill(2)}"
        )
        return _add_neg_to_rep(self._value, timecode)

    @property
    def frames(self) -> int:
        """
        frames returns the frame number of this timecode (how many frames would have
        played starting at 0 between 00:00:00:00 and the tc this value represents.).
        """
        return _rational_to_frames(self._value, self._rate)

    @property
    def feet_and_frames(self) -> str:
        """
        returns the number of feet and frames this timecode represents if it were
        shot on 35mm 4-perf film (16 frames per foot). ex: '5400+13'.

        feet and frames is most commonly used as a reference in the sound mixing world.
        """
        feet, frames = divmod(abs(self.frames), _FRAMES_PER_FOOT)
        feet_and_frames = f"{feet}+{str(frames).zfill(2)}"
        return _add_neg_to_rep(self._value, feet_and_frames)

    @property
    def seconds(self) -> decimal.Decimal:
        """
        seconds returns the number of seconds that would have elapsed between
        00:00:00:00 and the timecode this value represents.

        This value is a decimal.Decimal value to avoid floating-point shenanigans.
        """
        return decimal.Decimal(self._value.numerator) / decimal.Decimal(
            self._value.denominator
        )

    @property
    def premiere_ticks(self) -> PremiereTicks:
        """
        premiere_ticks returns the number of elapsed ticks this timecode represents in
        Adobe Premiere Pro.

        Premiere uses ticks internally to track elapsed time. A second contains
        254016000000 ticks, regardless of framerate.

        These ticks are present in Premiere FCP7XML cutlists, and can sometimes be used
        for more precise calculations during respeeds.

        Ticks are also used for scripting in Premiere Panels.
        """
        return PremiereTicks(round(self._value * _PPRO_TICKS_PER_SECOND))

    def runtime(self, precision: Optional[int] = 9) -> str:
        """
        Runtime returns the true runtime of the timecode in HH:MM:SS.FFFFFFFFF format.

        :param precision: how many places to print for fractional seconds.
            None=no rounding.

        Runtime will always be returned with at least one decimal place in order to
        distinguish it from timecode without an hours value. A runtime of exactly one
        hour will be returned as '01:00:00.0'

        Note: The true runtime will often diverge from the hours, minutes, and seconds
            value of the timecode representation when dealing with non-whole-frame
            framerates. Even drop-frame timecode does not continuously adhere 1:1 to the
            actual runtime. For instance, [01:00:00;00 @ 29.97 DF] has a true runtime of
            '00:59:59.9964', and [01:00:00:00 @ 23.98 NTSC] has a true runtime of
            '01:00:03.6'
        """
        seconds = round(abs(self.seconds), ndigits=precision)

        hours, seconds = divmod(seconds, _SECONDS_PER_HOUR)
        minutes, seconds = divmod(seconds, _SECONDS_PER_MINUTE)
        seconds, fractal = divmod(seconds, 1)
        if fractal == 0:
            fractal_str = ".0"
        else:
            fractal_str = "." + str(fractal).split(".")[-1].rstrip("0")

        runtime = (
            f"{str(hours).zfill(2)}:{str(minutes).zfill(2)}:"
            f"{str(seconds).zfill(2)}{fractal_str}"
        )

        return _add_neg_to_rep(self._value, runtime)

    def rebase(self, new_rate: FramerateSource) -> "Timecode":
        """
        rebase re-calculates the timecode at a new frame rate based on the frame-count
        value of the current timecode.

        :param new_rate: the new rate to rebase at.

        :returns: The new, rebased timecode.
        """
        return Timecode(self.frames, rate=new_rate)


def _add_neg_to_rep(frac_val: fractions.Fraction, rep: str) -> str:
    """
    _add_neg_to_rep adds a negative sign to a string tc representation if the value is
    less than 0.
    """
    if frac_val >= 0:
        return rep

    return "-" + rep


def _coerce_other(other: TimecodeSource, this_rate: Framerate) -> Timecode:
    """
    coerce other coerces a value to a timecode that an existing timecode needs to do
    an operation with. If the other value is not a timecode, the rate of the ecisting
    timecode will be used in parsing the other value.
    """
    if isinstance(other, Timecode):
        return other

    return Timecode(other, rate=this_rate)



import fractions
import decimal
import re
from typing import Union, List

from ._framerate import Framerate
from ._premiere_ticks import PremiereTicks
from ._timecode_sections import TimecodeSections
from ._timecode_dropframe import _parse_drop_frame_adjustment
from ._consts import (
    _tc_regex,
    _runtime_regex,
    _feet_and_frames_regex,
    _SECONDS_PER_MINUTE,
    _SECONDS_PER_HOUR,
    _FRAMES_PER_FOOT,
    _PPRO_TICKS_PER_SECOND,
)


# Union type detailing what types a Timecode can be parsed from.
_TimecodeParseSource = Union[
    str,
    int,
    float,
    fractions.Fraction,
    decimal.Decimal,
    PremiereTicks,
]


def _parse(src: _TimecodeParseSource, rate: Framerate) -> fractions.Fraction:
    """_parse converts an input value and rate into rational time."""
    if isinstance(src, str):
        return _parse_str(src, rate)
    # Premiere ticks check needs to come before int since it inherits int.
    elif isinstance(src, PremiereTicks):
        return _parse_premiere_ticks(src, rate)
    elif isinstance(src, int):
        return _parse_int(src, rate)
    elif isinstance(src, float):
        return _parse_float(src, rate)
    elif isinstance(src, decimal.Decimal):
        return _parse_decimal(src, rate)
    elif isinstance(src, fractions.Fraction):
        return _parse_fraction(src, rate)
    else:
        raise TypeError(f"unsupported type for Timecode conversion: {type(src)}")


def _parse_str(src: str, rate: Framerate) -> fractions.Fraction:
    """
    parse non-numeric string parses a string that represents a timecode, runtime or
    feet+frames.
    """
    # Match against our tc regex.
    matched = _tc_regex.fullmatch(src)
    if matched:
        return _parse_tc_str(matched, rate)

    matched = _runtime_regex.fullmatch(src)
    if matched:
        return _parse_runtime_str(matched, rate)

    matched = _feet_and_frames_regex.fullmatch(src)
    if matched:
        return _parse_feet_and_frames_str(matched, rate)

    # If there is no match, raise a regex error.
    raise ValueError(f"{repr(src)} is not a recognized timecode format")


def _parse_tc_str(matched: re.Match, rate: Framerate) -> fractions.Fraction:
    """
    _parse_tc_str converts a timecode string (ex: 01:00:00:00) into a fractional
    seconds value.
    """
    # We will always have a 'frames' group.
    frames: int = int(matched.group("frames"))
    # Some timecodes may be abbreviated, so the other three sections may or may not be
    # present. Gather them then filter out empty sections.
    groups: List[int] = [
        matched.group("section1"),
        matched.group("section2"),
        matched.group("section3"),
    ]
    groups = [x for x in groups if x]

    # If we got a hit on the negative group, this value is negative.
    is_negative = matched.group("negative") is not None

    # Work backwards to fill in the sections that are present, otherwise the value is
    # '0'.
    seconds = 0
    minutes = 0
    hours = 0
    if len(groups) >= 1:
        seconds = int(groups[-1])
    if len(groups) >= 2:
        minutes = int(groups[-2])
    if len(groups) >= 3:
        hours = int(groups[-3])

    if rate.ntsc:
        calc_rate = Framerate(round(rate.playback), ntsc=False)
    else:
        calc_rate = rate

    drop_adjustment = fractions.Fraction(0, 1)
    if rate.dropframe:
        drop_adjustment = _parse_drop_frame_adjustment(
            sections=TimecodeSections(
                negative=is_negative,
                hours=hours,
                minutes=minutes,
                seconds=seconds,
                frames=frames,
            ),
            rate=calc_rate,
        )

    # Divide the frames by the rate then add to the seconds value.
    seconds = seconds + minutes * _SECONDS_PER_MINUTE + hours * _SECONDS_PER_HOUR

    # Now get the frames as a fractional and then convert to an int.
    frames_frac = frames + seconds * calc_rate.playback

    # Add the drop-frame drop_adjustment
    frames_frac += drop_adjustment

    if is_negative:
        frames_frac = -frames_frac

    # Now parse the frame count.
    return _parse_int(round(frames_frac), rate)


def _parse_runtime_str(matched: re.Match, rate: Framerate) -> fractions.Fraction:
    """
    _parse_runtime_str parses a runtime string (ex 01:00:00.6) into a fractional
    seconds value.
    """
    # We will always have a 'frames' group.
    seconds: decimal.Decimal = decimal.Decimal(matched.group("seconds"))
    # Some timecodes may be abbreviated, so the other three sections may or may not be
    # present. Gather them then filter out empty sections.
    groups: List[int] = [
        matched.group("section1"),
        matched.group("section2"),
    ]

    groups = [x for x in groups if x]

    # Work backwards to fill in the sections that are present, otherwise the value is
    # '0'.
    minutes = 0
    hours = 0
    if len(groups) >= 1:
        minutes = int(groups[-1])
    if len(groups) >= 2:
        hours = int(groups[-2])

    # Calculate the number of seconds as a decimal, then use our decimal parser to
    # arrive at the correct result.
    seconds = seconds + minutes * _SECONDS_PER_MINUTE + hours * _SECONDS_PER_HOUR

    if matched.group("negative"):
        seconds = -seconds

    return _parse_decimal(seconds, rate)


def _parse_feet_and_frames_str(
    matched: re.Match,
    rate: Framerate,
) -> fractions.Fraction:
    """
    _parse_runtime_str parses a feet+frames string (ex 5400+13) into a fractional
    seconds value.
    """
    feet_str = matched.group("feet")
    frames_str = matched.group("frames")

    # Get the total frame count.
    frames = int(feet_str) * _FRAMES_PER_FOOT + int(frames_str)

    if matched.group("negative"):
        frames = -frames

    # Pass it off to our frame count parser.
    return _parse_int(frames, rate)


def _parse_int(frames: int, rate: Framerate) -> fractions.Fraction:
    """_parse_int converts the frame count into rational time."""
    return frames / rate.playback


def _parse_float(seconds: float, rate: Framerate) -> fractions.Fraction:
    """_parse_float converts the seconds value into rational time."""
    frames = _rational_to_frames(fractions.Fraction(seconds), rate)
    return _parse_int(frames, rate)


def _parse_fraction(seconds: fractions.Fraction, rate: Framerate) -> fractions.Fraction:
    """
    _parse_fraction parses a fractional seconds value to a fractional frames value.
    """
    return _parse_int(_rational_to_frames(seconds, rate), rate)


def _parse_decimal(seconds: decimal.Decimal, rate: Framerate) -> fractions.Fraction:
    """
    _parse_decimal parses a decimal.Decimal seconds value to a frame count fractional
    value.
    """
    frames = _rational_to_frames(fractions.Fraction(seconds), rate)
    return _parse_int(frames, rate)


def _parse_premiere_ticks(ticks: PremiereTicks, rate: Framerate) -> fractions.Fraction:
    seconds = ticks / _PPRO_TICKS_PER_SECOND
    return _parse_fraction(seconds, rate)


def _rational_to_frames(seconds: fractions.Fraction, rate: Framerate) -> int:
    """_rational_to_frames converts rational seconds to a frame count."""
    # multiply the fraction to frames.
    frac_frames = seconds * rate.playback

    # If this value converted cleanly, we can just return the numerator.
    if frac_frames.denominator == 1:
        return frac_frames.numerator

    # Otherwise lets convert to a float, then round.
    return round(float(frac_frames))



# "noqa" setting stops flake8 from flagging unused imports in __init__

from ._version import __version__  # noqa

from ._framerate import Framerate, FramerateSource, RATE
from ._timecode import Timecode, TimecodeSource
from ._premiere_ticks import PremiereTicks


(Framerate, FramerateSource, Timecode, TimecodeSource, RATE, PremiereTicks)



import fractions
import re

# _tc_regex is the regex pattern for parsing a timecode string (ex: '01:00:00:00').
_tc_regex: re.Pattern = re.compile(
    r"(?P<negative>-)?"
    r"((?P<section1>[0-9]+)[:|;])?((?P<section2>[0-9]+)[:|;])?"
    r"((?P<section3>[0-9]+)[:|;])?(?P<frames>[0-9]+)$"
)

# _runtime_regex is the regex pattern for parsing a runtime string (ex: '01:00:00.0')
_runtime_regex: re.Pattern = re.compile(
    r"(?P<negative>-)?"
    r"((?P<section1>[0-9]+)[:|;])?((?P<section2>[0-9]+)[:|;])?"
    r"(?P<seconds>[0-9]+(\.[0-9]+)?)$"
)

# _feet_and_frames_regex is the regex patter for matching a feet and frames string
# (ex: 5400+13)
_feet_and_frames_regex: re.Pattern = re.compile(
    r"(?P<negative>-)?(?P<feet>[0-9]+)\+(?P<frames>[0-9]+)",
)

# Cache the number of seconds in a minute and hour.
_SECONDS_PER_MINUTE: int = 60
_SECONDS_PER_HOUR: int = 3600

# _PPRO_TICKS_PER_SECOND is the number of Adobe Premiere Pro ticks in a second. This
# value is constant regardless of framerate.
_PPRO_TICKS_PER_SECOND: fractions.Fraction = fractions.Fraction(254016000000, 1)

# _FRAMES_PER_FOOT is the number of frames in a foot of 35mm, 4-perf film.
_FRAMES_PER_FOOT: int = 16



from typing import Union, Tuple


class PremiereTicks(int):

    """
    PremiereTicks signals than an int value represents an Adobe Premier Pro ticks value
    and can be used to wrap ints for parsing tc and doing mathematical operations.
    """

    # We need to override all of the mathematical magic methods or math operations,
    # even on two PremiereTicks instances, will return a generic int.
    def __add__(self, other: Union["PremiereTicks", int]) -> "PremiereTicks":
        """__add__ applies the int operation then casts to the correct type."""
        res = super().__add__(other)
        return self.__class__(res)

    def __radd__(self, other: Union["PremiereTicks", int]) -> "PremiereTicks":
        """__radd__ applies the int operation then casts to the correct type."""
        res = super().__radd__(other)
        return self.__class__(res)

    def __sub__(self, other: Union["PremiereTicks", int]) -> "PremiereTicks":
        """__sub__ applies the int operation then casts to the correct type."""
        res = super().__sub__(other)
        return self.__class__(res)

    def __rsub__(self, other: Union["PremiereTicks", int]) -> "PremiereTicks":
        """__rsub__ applies the int operation then casts to the correct type."""
        res = super().__rsub__(other)
        return self.__class__(res)

    def __mul__(self, other: Union["PremiereTicks", int]) -> "PremiereTicks":
        """__mul__ applies the int operation then casts to the correct type."""
        res = super().__mul__(other)
        return self.__class__(res)

    def __rmul__(self, other: Union["PremiereTicks", int]) -> "PremiereTicks":
        """__rmul__ applies the int operation then casts to the correct type."""
        res = super().__rmul__(other)
        return self.__class__(res)

    def __floordiv__(self, other: Union["PremiereTicks", int]) -> "PremiereTicks":
        """__floordiv__ applies the int operation then casts to the correct type."""
        res = super().__floordiv__(other)
        return self.__class__(res)

    def __rfloordiv__(self, other: Union["PremiereTicks", int]) -> "PremiereTicks":
        """__rfloordiv__ applies the int operation then casts to the correct type."""
        res = super().__rfloordiv__(other)
        return self.__class__(res)

    def __mod__(self, other: Union["PremiereTicks", int]) -> "PremiereTicks":
        """__mod__ applies the int operation then casts to the correct type."""
        res = super().__mod__(other)
        return self.__class__(res)

    def __rmod__(self, other: Union["PremiereTicks", int]) -> "PremiereTicks":
        """__rmod__ applies the int operation then casts to the correct type."""
        res = super().__rmod__(other)
        return self.__class__(res)

    def __divmod__(
        self,
        other: Union["PremiereTicks", int],
    ) -> Tuple["PremiereTicks", "PremiereTicks"]:
        """__divmod__ applies the int operation then casts to the correct type."""
        res = super().__divmod__(other)
        return self.__class__(res[0]), self.__class__(res[1])

    def __rdivmod__(
        self,
        other: Union["PremiereTicks", int],
    ) -> Tuple["PremiereTicks", "PremiereTicks"]:
        """__rdivmod__ applies the int operation then casts to the correct type."""
        res = super().__rdivmod__(other)
        return self.__class__(res[0]), self.__class__(res[1])

    def __floor__(self) -> "PremiereTicks":
        """__floor__ applies the int operation then casts to the correct type."""
        res = super().__floor__()
        return self.__class__(res)

    def __ceil__(self) -> "PremiereTicks":
        """__ceil__ applies the int operation then casts to the correct type."""
        res = super().__ceil__()
        return self.__class__(res)

    def __neg__(self) -> "PremiereTicks":
        """__neg__ applies the int operation then casts to the correct type."""
        res = super().__neg__()
        return self.__class__(res)

    def __abs__(self) -> "PremiereTicks":
        """__abs__ applies the int operation then casts to the correct type."""
        res = super().__abs__()
        return self.__class__(res)



import fractions

from ._framerate import Framerate
from ._timecode_sections import TimecodeSections


def _parse_drop_frame_adjustment(
    sections: TimecodeSections,
    rate: Framerate,
) -> fractions.Fraction:
    """
    _parse_drop_frame adjusts the frame number based on drop-frame TC conventions.

    Algorithm adapted from:
    https://www.davidheidelberger.com/2010/06/10/drop-frame-timecode/
    """
    timebase = rate.timebase
    drop_frames = round(timebase * 0.066666)

    # We have a bad frame value if our 'frames' place is less than the drop_frames we
    # skip on minutes not divisible by 10.
    has_bad_frame = sections.frames < drop_frames
    is_tenth_minute = sections.minutes % 10 == 0 or sections.minutes == 0
    if has_bad_frame and not is_tenth_minute:
        raise ValueError(
            f"drop-frame tc cannot have a frames value of less than {drop_frames} on "
            f"minutes not divisible by 10, found '{sections.frames}'",
        )

    total_minutes = 60 * sections.hours + sections.minutes
    adjustment = drop_frames * (total_minutes - total_minutes // 10)

    return -fractions.Fraction(adjustment, 1)


def _frame_num_to_drop_frame_num(
    frame_number: int,
    timebase: fractions.Fraction,
) -> int:
    """
    _frame_num_to_drop_frame_num converts a frame-number to an adjusted frame number for
    creating drop-frame tc.

    Algorithm adapted from:
    https://www.davidheidelberger.com/2010/06/10/drop-frame-timecode/

    :param frame_number: the frame number to convert to a drop-frame number.
    :param timebase: the timebase (not playback) to use for the conversion.

    :returns: The frame number adjusted to produce the correct drop-frame timecode when
    used in the normal timecode calculation.
    """
    # Get the number frames-per-minute at the whole-frame rate
    frames_per_minute_whole = timebase * 60
    # Get the number of frames we need to drop each time we drop frames (ex: 2 or 29.97)
    drop_frames = round(timebase * 0.066666)

    # Get the number of frames are in a minute where we have dropped frames at the
    # beginning
    frames_per_minute_drop = (timebase * 60) - drop_frames
    # Get the number of actual frames in a 10-minute span for drop frame timecode. Since
    # we drop 9 times a minute, it will be 9 drop-minute frame counts + 1 whole-minute
    # frame count.
    frames_per_10minutes_drop = frames_per_minute_drop * 9 + frames_per_minute_whole

    # Get the number of 10s of minutes in this count, and the remaining frames.
    tens_of_minutes, frames = divmod(frame_number, frames_per_10minutes_drop)

    # Create an adjustment for the number of 10s of minutes. It will be 9 times the
    # drop value (we drop for the first 9 minutes, then leave the 10th alone).
    adjustment = 9 * drop_frames * tens_of_minutes

    # If our remaining frames are less than a whole minute, we aren't going to drop
    # again. Add the adjustment and return.
    if frames < frames_per_minute_whole:
        return frame_number + adjustment

    # Remove the first full minute (we don't drop until the next minute) and add the
    # drop-rate to the adjustment.
    frames -= timebase
    adjustment += drop_frames

    # Get the number of remaining drop-minutes present, and add a drop adjustment for
    # each.
    minutes_drop = frames // frames_per_minute_drop
    adjustment += minutes_drop * drop_frames

    # Return our original frame number adjusted by our calculated adjustment.
    return frame_number + adjustment



from typing import NamedTuple


class TimecodeSections(NamedTuple):
    """
    TimecodeSections contains the sections of a parsed timecode value as ints.
    """

    # True if this is a negative timecode. All other values will always be positive.
    negative: bool
    hours: int
    minutes: int
    seconds: int
    frames: int



import dataclasses
import fractions
from typing import Union, Tuple, Optional


class Framerate:
    def __init__(
        self,
        src: "FramerateSource",
        *,
        ntsc: Optional[bool] = None,
        dropframe: bool = False,
    ) -> None:
        """
        Framerate is the rate at which a video file frames are played back.

        Framerate is measured in frames-per-second (24/1 = 24 frames-per-second).

        :param src: the source value to parse to a Framerate. src must conform to
            one of the following value types:

            - :class:`Framerate`: Another timebase value.

            - ``str``: Strings may be of a fractional value ('24/1'), whole numbers
              representing the fps ('24') OR decimal numbers like ('23.9876') or 23.98.
              Such values will be converted to their actual meaning (24000/1001).

            - ``Tuple[int, int]``: Numerator and denominator of a fraction.

            - ``fractions.Fraction`` value.

            - ``float`` value -- only allowed if ntsc=true.

            .. note ::

                if a value like 1/24 is passed in (where the numerator is less than
                the denominator), the value will be converted to 24/1 after being
                parsed.

        :param ntsc: Whether timecode should be calculated via NTSC convention.

            For NTSC, playback speed is 23.976, the TIMECODE is still calculated as if
            it were at 24fps, which affects how a :func:`~Timecode.timecode` value
            will be rendered.

            When ``None``, non-whole number floats like 23.98 and fractional values
            with a denominator of 1001 will be assumed to be ntsc, but whole-numbers
            will be left as-is.

            When this option is ``True``, the framerate will be rounded to the nearest
            whole number before the :func:`~Timecode.timecode` representation is
            calculated.

        :param dropframe: Whether this is a drop-frame style timecode (only available
            for frame rates divisible by 30000/1001 like 29.97 and 59.94).
        """
        dropframe, ntsc = _validate_dropframe_ntsc(ntsc, dropframe)

        self._value: fractions.Fraction
        self._dropframe = dropframe
        self._ntsc: bool

        # Parse tha value into a timebase.
        self._value = _parse(src, ntsc)
        if isinstance(src, Framerate):
            self._dropframe = src.dropframe
            ntsc = src.ntsc

        self._ntsc = _infer_ntsc(self._value, ntsc)
        _validate_drop_frame_value(self._value, self._dropframe)

    def __str__(self) -> str:
        """Returns the framerate as a fractional string (ex: '24/1')."""
        return str(self._value)

    def __repr__(self) -> str:
        """Returns formatted framerate information: (ex: '[24/1 fps NTSC]')."""
        rate_value = str(round(float(self._value), 2)).rstrip("0").rstrip(".")
        value = f"[{rate_value}"
        if self._ntsc:
            value += " NTSC"

        if self._dropframe:
            value += " DF"

        value += "]"

        return value

    def __eq__(self, other: object) -> bool:
        """
        Two Framerate instances are equal if their fractional value, ntsc attribute and
        drop frame attribute are equal.
        """
        if not isinstance(other, Framerate):
            return NotImplemented

        return (
            self._value == other._value
            and self._ntsc == other._ntsc
            and self._dropframe == other._dropframe
        )

    @property
    def playback(self) -> fractions.Fraction:
        """
        The rational representation of the real-world video playback speed in
        frames-per-second as a fraction.

        example: 24000/1001.

        .. note::
            :func:`~Framerate.playback` and :func:`~Framerate.timebase` only differ in
            NTSC framerates. All non-NTSC framerates will have identical values for
            both properties.
        """
        return self._value

    @property
    def timebase(self) -> fractions.Fraction:
        """
        The rational representation of the timecode display speed in frames-per-second
        as a fraction.

        example: 24/1

        .. note::
            :func:`~Framerate.playback` and :func:`~Framerate.timebase` only differ in
            NTSC framerates. All non-NTSC framerates will have identical values for
            both properties.
        """
        # If this is an NTSC timebase, convert to a rounded value over 1.
        if self._ntsc:
            return fractions.Fraction(round(self._value), 1)

        # Otherwise return our internal value.
        return self._value

    @property
    def ntsc(self) -> bool:
        """Whether this is an NTSC-style time base (aka 23.98, 24000/1001, etc)."""
        return self._ntsc

    @property
    def dropframe(self) -> bool:
        """Whether this Framerate is drop-frame."""
        return self._dropframe


def _validate_dropframe_ntsc(
    ntsc: Optional[bool],
    dropframe: bool,
) -> Tuple[bool, Optional[bool]]:
    """
    _validate_dropframe_ntsc validates that the drop frame and ntsc arguments are not
    in conflict and returns adjusted ones.
    """
    if dropframe:
        # If NTSC was explicitly set to False and dropframe was explicitly set to true
        # then there is a conflict.
        if ntsc is False:
            raise ValueError(
                "ntsc must be [True] or [None] if drop_frame is [True]",
            )

        # Otherwise if dropframe is true, NTSC must be set to true as well.
        ntsc = True

    return dropframe, ntsc


def _infer_ntsc(value: fractions.Fraction, ntsc: Optional[bool]) -> bool:
    """
    _infer_ntsc looks at the parsed fraction value and infers if it should be considered
    NTSC.
    """
    # If no explicit ntsc value was set, assume that values with a denominator of
    # 1001 are ntsc.
    if ntsc is None:
        if value.denominator == 1001:
            ntsc = True
        else:
            ntsc = False
    else:
        ntsc = ntsc

    return ntsc


def _validate_drop_frame_value(value: fractions.Fraction, dropframe: bool) -> None:
    """
    _validate_drop_frame_value validates that a framerate value with dropframe enabled
    is a multiple of 30000/1001.
    """
    # Validate that drop-frame TC is cleanly divisible by 30000/1001. Drop-frame is
    # not defined for any other timebases. Generally it is only allowed for 29.97
    # and 59.94
    if dropframe and value % fractions.Fraction(30000, 1001) != 0:
        raise ValueError(
            "drop_frame may only be true if framerate is divisible by "
            "30000/1001 (29.97)"
        )


def _parse(src: "FramerateSource", ntsc: Optional[bool]) -> fractions.Fraction:
    """_parse does the heavy lifting of parsing a value into a TimeBase."""
    if isinstance(src, str):
        frac = _parse_string(src, ntsc)
    elif isinstance(src, int):
        frac = fractions.Fraction(src, 1)
    elif isinstance(src, float):
        frac = _parse_float(src, ntsc)
    elif isinstance(src, tuple):
        if len(src) > 2:
            raise ValueError(
                f"Framerate tuple value must contain exactly 2 values, got "
                f"{len(src)}",
            )
        frac = fractions.Fraction(numerator=src[0], denominator=src[1])
    elif isinstance(src, Framerate):
        frac = src.playback
    elif isinstance(src, fractions.Fraction):
        frac = src
    else:
        raise TypeError(f"unsupported type for Framerate conversion: {type(src)}")

    # If the numerator is less than the denominator, swap the values.
    if frac.numerator < frac.denominator:
        frac = fractions.Fraction(
            numerator=frac.denominator,
            denominator=frac.numerator,
        )

    if ntsc:
        # If this is an ntsc timebase, we need to round up what we were given to the
        # nearest whole number and divide by 1001.
        frac = fractions.Fraction(round(frac) * 1000, 1001)

    return frac


def _parse_string(src: str, ntsc: Optional[bool]) -> fractions.Fraction:
    """
    _parse_string parses a string value to a fraction.Fraction and drop frame
    boolean.
    """
    # If this looks like a fraction (has a '/'), try to parse it as such.
    if "/" in src:
        return fractions.Fraction(src)

    # If this is a whole number like "24.0", we want to just strip it.
    if "." in src:
        try:
            src_float = float(src)
        except ValueError:
            pass
        else:
            return _parse_float(src_float, ntsc)

    # Try to parse as an int next.
    try:
        frac_int = int(src)
    except ValueError:
        raise ValueError(f"could not parse Framerate value of {repr(src)}")
    # If this int passed, return 1/[value] as the TimeBase
    return fractions.Fraction(numerator=1, denominator=frac_int)


def _parse_float(src: float, ntsc: Optional[bool]) -> fractions.Fraction:
    if not src.is_integer() and ntsc is False:
        raise ValueError(
            "non-whole-number floats values cannot be parsed when ntsc=False. "
            "use precise fraction.Fraction value instead",
        )

    # If this is an integer, returns [src]/1
    if src.is_integer():
        return fractions.Fraction(int(src))

    # Otherwise we are going to assume this in an NTSC-style rate (like 29.97), and
    # round up to the nearest whole number and use a denominator of 1001.
    return fractions.Fraction(round(src) * 1000, 1001)


FramerateSource = Union[Framerate, str, Tuple[int, int], fractions.Fraction, float]
"""FramerateSource is the set of types a Framerate can be created from"""


# We'll use a frozen dataclass for this so the values cannot be changed.
@dataclasses.dataclass(frozen=True)
class _Rates:
    """
    _Rates is used as a one-off na to hold a number of common pre-defined framerates for
    callers to use.
    """

    # 23.98 fps NTSC.
    F23_98: Framerate = Framerate(23.98, ntsc=True)

    # 24 fps.
    F24: Framerate = Framerate(24)

    # 29.97 fps NTSC.
    F29_97_NDF: Framerate = Framerate(29.97, ntsc=True)

    # 29.97 fps DROP FRAME.
    F29_97_DF: Framerate = Framerate(29.97, dropframe=True)

    # 30 fps NTSC.
    F30: Framerate = Framerate(30)

    # 47.95 fps NTSC.
    F47_95: Framerate = Framerate(47.95, ntsc=True)

    # 48 fps NTSC.
    F48: Framerate = Framerate(48)

    # 59.94 fps NTSC.
    F59_94_NDF: Framerate = Framerate(59.94, ntsc=True)

    # 59.94 fps NTSC DROP FRAME.
    F59_94_DF: Framerate = Framerate(59.94, dropframe=True)

    # 60 fps NTSC.
    F60: Framerate = Framerate(60)


RATE: _Rates = _Rates()
"""RATE holds a number of pre-defined frame rates for convenience"""
